from contextlib import suppress
from math import log2
from re import findall

from kalib.descriptors import Property
from kalib.functions import to_ascii
from kalib.internals import Who, unique
from kalib.misc import lazy_proxy_to


def generate_offsets(x):
    power = int(log2(x))

    yield 0
    for base in range(power):
        bias = (2 ** (power - base))
        for no in range(x // bias + 1):
            if no % 2:
                yield no * bias


def replace_class_with_str(method, *args, **kw):
    if args and isinstance(args[0], Str):
        args = (str(args[0]), *args[1:])
    return method(*args, **kw)


def get_mime(data):
    from kalib.importer import optional
    if (
        data and
        (read := optional('magic.from_buffer', quiet=False)) and
        (detect_from := optional('magic.detect_from_content'))
    ):
        return {'text': read(data), 'type': detect_from(data).mime_type}


@lazy_proxy_to(
    'string',
    '__add__', '__contains__', '__eq__', '__format__', '__hash__', '__ge__',
    '__getitem__', '__gt__', '__iter__', '__le__', '__len__', '__lt__', '__mod__',
    '__mul__', '__ne__', '__rmod__', '__rmul__', '__sizeof__', 'capitalize',
    'casefold', 'center', 'count', 'encode', 'endswith', 'expandtabs', 'find',
    'format', 'format_map', 'index', 'isalnum', 'isalpha', 'isascii', 'isdecimal',
    'isdigit', 'isidentifier', 'islower', 'isnumeric', 'isprintable', 'isspace',
    'istitle', 'isupper', 'join', 'ljust', 'lower', 'lstrip', 'maketrans',
    'partition', 'removeprefix', 'removesuffix', 'replace', 'rfind', 'rindex',
    'rjust', 'rpartition', 'rsplit', 'rstrip', 'split', 'splitlines', 'startswith',
    'strip', 'swapcase', 'title', 'translate', 'upper', 'zfill',
    Property.Cached, pre=replace_class_with_str,
)
class Str:
    """A class for handling bytes | str objects.

    Autodetect charsets, autoencode/decode from bytes/text.
    Just pass any bytes | str object to Str and use it:

    x = Str(b'hello')
    str(x) == 'hello'
    bytes(x) == b'hello'

    """

    INTERNAL_CHARSET = 'utf-8'
    CHARDET_MAX_READ = 2 ** 20

    to_ascii = staticmethod(to_ascii)

    @classmethod
    def uncast(cls, something):
        if isinstance(something, bytes | str):
            return something

        elif isinstance(something, int):
            return str(something)

        if isinstance(something, Exception):
            return f'{Who(something)}: {something}'

        msg = f"({Who(something)}) {something=} isn't valid bytes | str"
        raise TypeError(msg)

    @classmethod
    def to_bytes(cls, something, *args, **kw):
        return bytes(cls(something, *args, **kw))

    @classmethod
    def to_str(cls, something, *args, **kw):
        return str(cls(something, *args, **kw))

    def __init__(self, something, /, encoding=None):
        self._object = something
        self._encoding = encoding or self.INTERNAL_CHARSET

    @Property.Cached
    def representation(self):
        return self.uncast(self._object)

    @Property.Cached
    def mime(self):
        with suppress(ImportError):
            return get_mime(self.bytes)

    @Property.Cached
    def chardet(self):
        from kalib.importer import required
        with suppress(Exception):
            func = required('charset_normalizer.detect')
            return func(bytes(self))

    @Property.Cached
    def charset_probe_order(self):
        result = [self._encoding]
        if meta := self.chardet:
            result.append(meta['encoding'])
        return tuple(unique([*result, 'ascii']))

    @Property.Cached
    def charset(self):
        string = self.representation
        is_bytes = isinstance(string, bytes)

        def binary_iter(x):
            getter = x.__getitem__
            offset = len(x) - len(x) % 2 - 1
            if offset != -1:

                iterator = generate_offsets(offset)
                offsets = map(getter, iterator)
                if not is_bytes:
                    offsets = map(ord, offsets)
                yield from enumerate(offsets)

        limit = self.CHARDET_MAX_READ
        if not is_bytes and len(string) <= limit:
            limit = 0

        chars = []
        default = 'ascii'
        if not string:
            return default

        for no, char in binary_iter(string):
            if no > limit:
                break

            if char < 0x20:  # noqa: PLR2004
                if char in (0x9, 0xa, 0xc, 0xd):
                    continue
                return 'binary'

            if char >= 0x7f:  # noqa: PLR2004, ANSI Extended Border
                if is_bytes:
                    default = 'ansi'
                else:
                    chars.append(char)
                    if len(chars) > 2 ** 1024:
                        return self.charset_probe

        return default if is_bytes else self.charset_probe

    @Property.Cached
    def charset_probe(self):
        string = self.representation

        if isinstance(string, bytes):
            method = string.decode
            order = self.charset_probe_order

        else:
            method = string.encode
            order = tuple(reversed(self.charset_probe_order))

        for charset in filter(bool, order):
            try:
                method(charset)
                return charset  # noqa: TRY300

            except (UnicodeEncodeError, UnicodeDecodeError):
                ...

    @Property.Cached
    def bytes(self):
        string = self.representation
        return string if isinstance(string, bytes) else string.encode(self._encoding)

    @Property.Cached
    def string(self):
        string = self.representation
        if isinstance(string, str):
            return string

        charset = self.charset
        charset = self.charset_probe if charset == 'ansi' else charset

        if charset != 'binary':
            return string.decode(charset)

        msg = f"couldn't decode {string!r} ({string!a})"
        if meta := self.charset:
            msg = f'{msg}, meta: {meta}'
        raise ValueError(msg)

    def __bytes__(self):
        return self.bytes

    def __str__(self):
        return self.string

    def __repr__(self):
        string = self.representation
        size = f'{len(self.bytes):d}'

        length = ''
        charset = (f'{self.charset} '.upper()) if self.charset else ''

        if isinstance(string, str) and len(self.bytes) != len(self.string):
            length = f'={len(self.string):d}'

        return (
            f'<{charset}{Who(self, full=False)}'
            f'[{Who(self._object, full=False)}'
            f'({size})]{length} at {id(self):#x}>')

    def tokenize(self, regex=r'([\w\d]+)'):
        return findall(regex, self.string)



from asyncio import ensure_future, iscoroutinefunction
from contextlib import suppress
from functools import cached_property, lru_cache, partial, wraps
from inspect import iscoroutine, isfunction, ismethod
from time import time

from kalib.classes import Nothing
from kalib.internals import (
    Is,
    Who,
    class_of,
    get_attr,
    get_owner,
    is_class,
)

__all__ = 'Property', 'cache'


def cache(limit=None):

    function = partial(lru_cache, maxsize=None, typed=False)

    if isinstance(limit, classmethod | staticmethod):
        msg = f"can't wrap {Who.Is(limit)}, you must use @cache after it"
        raise TypeError(msg)

    for func in isfunction, iscoroutine, ismethod:
        if func(limit):
            return function()(limit)

    if limit is not None and (not isinstance(limit, float | int) or limit <= 0):
        msg = f'limit must be None or positive integer, not {Who.Is(limit)}'
        raise TypeError(msg)

    return function(maxsize=limit) if limit else function()


def call_descriptor(descriptor):
    if Is.subclass(descriptor, BaseProperty):
        return descriptor.call

    func = getattr(descriptor, 'fget', Nothing)
    if func is Nothing:
        head = f'expected descriptor derived from {Who(BaseProperty)}'

        if class_of(descriptor) is not cached_property:
            raise TypeError(f'{head}, but got {Who(descriptor)} instead')

        raise TypeError(
            f'{head}, but got {Who(descriptor)}, may be you use '
            f'@Property.Bind instead @Property.Cached?')

    return func


def parent_call(func):

    @wraps(func)
    def parent_caller(node, *args, **kw):
        try:
            desc = get_attr(
                class_of(node), func.__name__, exclude_self=True,
                index=func.__name__ not in class_of(node).__dict__)
            return func(node, call_descriptor(desc)(node, *args, **kw), *args, **kw)

        except RecursionError as e:
            raise RecursionError(
                f'{Who(node)}.{func.__name__} call real {Who(func)}, '
                f"couldn't reach parent descriptor; "
                f"maybe {Who(func)} it's mixin of {Who(node)}?") from e

    return parent_caller


class PropertyError(Exception):
    ...


class ContextFaultError(PropertyError):
    ...


class AttributeException(PropertyError):  # noqa: N818
    ...


def invokation_context_check(func):

    @wraps(func)
    def context(self, node, *args, **kw):
        klass = self.klass
        if (
            klass is not None and
            (node is None or klass != is_class(node))
        ):
            msg = (
                f'{Who(func)} exception, '
                f'{self.header_with_context(node)}, {node=}')

            if node is None and not klass:
                msg = f'{msg}; looks like as non-instance invokation'
            raise ContextFaultError(msg)

        return func(self, node, *args, **kw)
    return context


class BaseProperty:

    klass = False
    readonly = False

    def __init__(self, function):
        self.function = function

    @cached_property
    def name(self):
        return self.function.__name__

    @cached_property
    def is_data(self):
        return bool(hasattr(self, '__set__') or hasattr(self, '__delete__'))

    @cached_property
    def title(self):
        mode = 'mixed' if self.klass is None else ('instance', 'class')[self.klass]

        prefix = ('', 'data ')[self.is_data]
        return (
            f'{mode} {prefix}descriptor '
            f'{Who(self, addr=True)}'.strip())

    @cached_property
    def header(self):
        try:
            return f'{self.title}({self.function!a})'
        except Exception:  # noqa: BLE001
            return f'{self.title}({Who(self.function)})'

    def header_with_context(self, node):

        if node is None:
            mode = 'mixed' if self.klass is None else 'undefined'
        else:
            mode = ('instance', 'class')[is_class(node)]

        return f'{self.header} with {mode} ({Who(node, addr=True)}) call'

    @invokation_context_check
    def get_node(self, node):
        return node

    @invokation_context_check
    def call(self, node):
        try:
            value = self.function(node)
            return ensure_future(value) if iscoroutinefunction(self.function) else value

        except AttributeError as e:
            from kalib.exceptions import exception
            error = AttributeException(exception(e).message.rsplit(':', 1)[-1].strip())

            error.exception = e
            raise error from e

    def __str__(self):
        return f'<{self.header}>'

    def __repr__(self):
        return f'<{self.title}>'

    def __get__(self, instance, klass):
        if instance is None and self.klass is False:
            raise ContextFaultError(self.header_with_context(klass))

        return self.call((instance, klass)[self.klass])


class InheritedClass(BaseProperty):

    """By default class property will be used parent class.
    This class change behavior to last inherited child.
    """

    @invokation_context_check
    def get_node(self, node):
        return node

    @classmethod
    def make_from(cls, klass):
        """Make child-aware class from plain parent-based."""

        name = Who(klass, full=False)
        suffix = f'{"_" if name == name.lower() else ""}inherited'.capitalize()
        return type(f'{name}{suffix}', (cls, klass), {})


class Cached(BaseProperty):

    @classmethod
    def expired_by(cls, callback):
        # result = partial(cls, expired=callback)
        # result.func = callback
        return partial(cls, expired=callback)

    @classmethod
    def ttl(cls, expire: float):
        if not isinstance(expire, float | int):
            raise TypeError(f'expire must be float or int, not {Who.Is(expire)}')

        if expire <= 0:
            raise ValueError(f'expire must be positive number, not {expire!r}')

        def is_fresh(node, value=Nothing):  # noqa: ARG001
            return (value + expire > time()) if value else time()

        return cls.expired_by(is_fresh)

    def __init__(self, function, expired=None):
        super().__init__(function)
        self.expired = expired

    @invokation_context_check
    def get_cache(self, node):
        name = f'__{("instance", "class")[is_class(node)]}_memoized__'

        if hasattr(node, '__dict__'):
            with suppress(KeyError):
                return node.__dict__[name]

        cache = {}
        setattr(node, name, cache)
        return cache

    @invokation_context_check
    def call(self, obj):
        node = self.get_node(obj)

        with suppress(KeyError):

            if not self.expired:
                return self.get_cache(node)[self.name]

            value, stamp = self.get_cache(node)[self.name]
            if self.expired(node, stamp) is not False:
                return value

        return self.__set__(node, super().call(obj))

    @invokation_context_check
    def __set__(self, node, value):
        cache = self.get_cache(node)
        if not self.expired:
            cache[self.name] = value
        else:
            cache[self.name] = value, self.expired(node)
        return value

    @invokation_context_check
    def __delete__(self, node):
        cache = self.get_cache(node)
        with suppress(KeyError):
            del cache[self.name]


class ClassProperty(BaseProperty):
    klass = True

    @invokation_context_check
    def get_node(self, node):
        return get_owner(node, self.name) if is_class(node) else node


class MixedProperty(ClassProperty):
    klass = None

    def __get__(self, instance, klass):
        return self.call(instance or klass)


class ClassCachedProperty(ClassProperty, Cached):
    """Class-level cached property that passes the original class as the first
    positional argument and replaces the original data-descriptor."""


class MixedCachedProperty(MixedProperty, Cached):
    """Mixed-level cached property that replaces the original data-descriptor"""


class PreCachedProperty(MixedProperty, Cached):
    @invokation_context_check
    def __set__(self, node, value):
        if not Is.Class(node):
            return value
        return super().__set__(node, value)


class PostCachedProperty(MixedProperty, Cached):
    @invokation_context_check
    def __set__(self, node, value):
        if Is.Class(node):
            return value
        return super().__set__(node, value)


class Property(BaseProperty):
    """
    The Property class is a versatile descriptor used for managing attributes in a class.
    It includes various types of properties for both instances and classes.
    """

    """Replaces the data-descriptor with a calculated result,
    supporting asynchronous operations."""
    Cached = Cached

    """Just replaces the getter with a calculated result for simple cases."""
    Bind = cached_property

    """Passes the class as the first positional argument."""
    Class = ClassProperty

    """Passes the original class where the decorator is used, then replaces its
    own data-descriptor with a calculated result in the original parent."""
    Class.Parent = ClassCachedProperty

    """Passes the inherited class and sets the child class attribute
    with a calculated result."""
    Class.Cached = InheritedClass.make_from(ClassCachedProperty)

    """Can be used with either a class or an instance."""
    Mixed = MixedProperty

    """Replaces the data-descriptor with a calculated result for instances, but when
    used in a class context, passes the class and just returns the result."""
    Mixed.Cached = InheritedClass.make_from(MixedCachedProperty)

    Any = MixedProperty
    Any.Pre    = PreCachedProperty
    Any.Post   = PostCachedProperty
    Any.Cached = InheritedClass.make_from(PostCachedProperty)
    Any.Pre.Parent  = InheritedClass.make_from(PreCachedProperty)
    Any.Post.Parent = InheritedClass.make_from(PostCachedProperty)

    @classmethod
    def Custom(cls, callback):  # noqa: N802
        result = Cached.expired_by(callback)
        result.Class = cls.Class.Cached.expired_by(callback)
        return result


# obsolete interface

class pin(Cached):  # noqa: N801

    attr = Property.Bind

    # class-descriptor, but .root uses class what have
    # declared pin.root, instead last child
    root = Property.Class.Parent
    cls  = Property.Class.Cached

    # mixed-descriptor, do not cache when using in class context, but bind instead
    any  = Property.Mixed.Cached



from contextlib import suppress
from dataclasses import _FIELD, _MISSING_TYPE, Field, is_dataclass
from dataclasses import dataclass as base_dataclass
from datetime import date, datetime
from enum import IntFlag, auto
from operator import attrgetter, itemgetter
from pathlib import Path
from types import ModuleType, UnionType
from typing import (
    Any,
    ClassVar,
    Generic,
    Optional,
    Union,
    _UnionGenericAlias,
    get_args,
    get_origin,
)

from kalib.classes import Nothing
from kalib.datastructures import json, loads
from kalib.descriptors import Property, cache
from kalib.importer import optional, required, sort
from kalib.internals import (
    Is,
    Who,
    class_of,
    get_owner,
    is_callable,
    is_class,
    is_collection,
    is_iterable,
    iter_inheritance,
    unique,
)
from kalib.loggers import Args, Logging
from kalib.misc import sourcefile
from kalib.text import Str

REPR_LIMIT = 2 ** 14


def capitalize_snakes(something):
    return ''.join(map(str.capitalize, filter(bool, something.split('_'))))


class ConfigurationError(Exception):
    ...


class ConfigurationEvaluationError(ConfigurationError):
    ...


class ConfigurationSchemeError(ConfigurationError):
    ...


class ConfigurationSchemeMissingError(ConfigurationSchemeError):
    ...


class ConfigurationTypeError(ConfigurationSchemeError, TypeError):
    ...


def fields_cast(data):
    return ', '.join(f'({Who(data[k])}){k}' for k in sort(data))


Dict = dict | optional('immutabledict.immutabledict', default=dict)


@base_dataclass(kw_only=True)
class FieldDescriptions:
    name          : str
    type          : Any
    default       : Any
    factory       : Any
    typename      : str
    default_class : Any
    required      : bool


class dataclass(Logging.Mixin):  # noqa: N801

    def __init_subclass__(cls, *args, **kw):

        kw.update(getattr(cls, '__dataclass_options__', None) or {})
        kw.setdefault('eq', True)
        kw.setdefault('frozen', True)
        kw.setdefault('kw_only', True)

        self = base_dataclass(cls, *args, **kw)
        generic_initialize = self.__init__

        def __init__(self, *args, **kw):  # noqa: N807
            self.__dict__.update({
                '__instance_memoized__': {},
                '__non_model_items__': kw.pop('unused', {})})
            generic_initialize(self, *args, **kw)

        self.__init__ = __init__
        return self

    class Policy(IntFlag):

        Strict  = auto()
        Include = auto()
        Exclude = auto()

    policy = Policy.Strict

    @Property.Class.Cached
    def __fields__(cls):
        def make(x):
            if get_origin(x.type) in (Generic, Union, UnionType):
                name = ' | '.join(map(Who, get_args(x.type)))
            else:
                name = Who(x.type)

            factory = x.default_factory
            required = (
                {_MISSING_TYPE} ==
                set(map(class_of, (x.default, factory))))

            klass = class_of(x.default)
            return FieldDescriptions(
                name     = x.name,
                type     = x.type,
                typename = name,

                factory = factory,
                default = x.default,
                default_class = klass,
                required = required,
            )

        result = filter(
            lambda x: ClassVar not in (x.type, get_origin(x.type)),
            cls.__dataclass_fields__.values())
        return tuple(map(make, result))

    @Property.Class.Cached
    def __fields_dict__(cls):
        fields = {f.name: f for f in cls.__fields__}
        if not Is.subclass(cls, dataclass):
            parents = class_of(cls).__mro__
            msg = f'{Who(cls)} must be a dataclass, {parents=}'
            raise ConfigurationTypeError(msg)

        for key, field in fields.items():
            value = field.default
            if (
                field.default_class is not _MISSING_TYPE
                and not Is.subclass(value, field.type)
            ):
                config = json.repr(fields)
                msg = (
                    f'{Who(cls)}: ({field.typename}){key} '
                    f'invalid default value ({Who(value)}), {config=}')
                raise ConfigurationEvaluationError(msg)

        return fields

    @Property.Class.Cached
    def __mro_classes__(cls):
        result = []
        logging = class_of(Logging.Mixin).__name__
        skip = {BaseAutoClass, DataClassFlexible, Logging.Mixin, dataclass}

        for c in iter_inheritance(cls, exclude_stdlib=False):
            if c not in skip and c.__name__ != logging:
                result.append(c)
        return tuple(result)

    @classmethod
    def _process_extra_items(cls, data):
        extra = {}
        if data:
            policy = cls.policy
            if policy & cls.Policy.Exclude:
                extra = dict(data)
            else:
                msg = f'{Who(cls)}[{policy=}]: extra: {fields_cast(data)} passed'
                raise ConfigurationEvaluationError(msg)

        return {}, extra

    @classmethod
    def _preload(cls, data, *args, **kw):  # noqa: ARG003
        return data

    @classmethod
    def _post(cls, data, *args, **kw):  # noqa: ARG003
        return data

    @classmethod
    def _process(cls, data, *args, **kw):  # noqa: ARG003, PLR0912
        fields = cls.__fields_dict__

        if isinstance(data, Dict):
            data = dict(data)

        if isinstance(data, dict):
            for key, field in fields.items():

                if field.type is Any:
                    data[key] = data.pop(key)
                    continue

                try:
                    if (
                        class_of(field.type) not in (UnionType, _UnionGenericAlias)
                        and Is.subclass(field.type, dataclass)
                        and isinstance(data[key], dataclass | dict)
                    ):
                        data[key] = field.type.load(data.pop(key))

                except KeyError as e:
                    msg = (
                        f'{Who(cls)}: ({field.typename}){key} '
                        f'required, got {fields_cast(data)}')
                    raise ConfigurationEvaluationError(msg) from e

            return data

        if not args:
            if is_dataclass(data):
                result = {}
                for key, field in fields.items():
                    try:
                        result[key] = getattr(data, key)
                    except AttributeError as e:
                        msg = (
                            f'{Who(cls)}: ({field.typename}){key} '
                            f'required, got {fields_cast(data)}')
                        raise ConfigurationEvaluationError(msg) from e

                return result

        elif len(args) + 1 != len(fields):
            data = json.repr(data)
            msg = (f'{Who(cls)}: scheme {len(fields):d} '
                   f"fields ({' '.join(tuple(fields))}) "
                   f"couldn't match to {len(args) + 1:d} "
                   f'taken args: {Args(data, *args)[:1024]}')
            raise ConfigurationEvaluationError(msg)

        args = (data, *args)
        if not fields and args:
            msg = (f"{Who(cls)}: scheme haven't any fields, "
                   f'taken args: {Args(args)}')
            raise ConfigurationEvaluationError(msg)

        required = len(list(filter(attrgetter('required'), fields.values())))
        if required > len(args):
            if isinstance(data, dict) and not args[1:]:
                if not set(data) - set(fields):
                    return data

                msg = f"taken data ({' '.join(data)})"
            else:
                msg = f'taken args: {Args(data, *args)}'

            msg = (
                f'{Who(cls)}: scheme {required=} '
                f"fields ({' '.join(tuple(fields))}) "
                f"couldn't match to {len(data):d}; {msg}")
            raise ConfigurationEvaluationError(msg)

        if not required and isinstance(data, dict) and not data:
            return {}

        return dict(zip(list(fields)[:len(args)], args, strict=True))

    @classmethod
    def from_file(cls, path, *args, **kw):
        with Path(path).open('rb') as fd:
            return cls.load(json.loads(fd.read()), *args, **kw)

    @classmethod
    def loads(cls, body, *args, **kw):
        return cls.load(loads(body) or json.loads(body), *args, **kw)

    @classmethod
    def load(cls, data, *args, **kw):
        if isinstance(data, cls) and class_of(data) is cls:
            return data.copy(**kw)

        fields = cls.__fields_dict__
        validate = kw.pop('validate', True)

        data = cls._preload(data, *args, **kw)
        data = cls._process(data, *args, **kw)

        additional_items, unused_items = (
            cls._process_extra_items(
                dict(unique(data, exclude=fields))))

        result = dict(unique(data, include=fields))
        result.update(additional_items)

        for key, field in unique(fields, exclude=result):
            value = field.default

            if Is.subclass(value, _MISSING_TYPE):
                types = getattr(field.type, '__args__', None)
                if (
                    (field.typename != 'builtins.NoneType') and
                    (not types or not Is.subclass(None, types))
                ):
                    if not Is.subclass(field.factory, _MISSING_TYPE):
                        value = field.factory()
                    else:
                        msg = (
                            f'{Who(cls)}: ({field.typename}){key} '
                            f'missing, got: {fields_cast(data)}')
                        raise ConfigurationEvaluationError(msg)
                else:
                    value = None

            result[key] = value

        self = cls(unused=unused_items, **cls._post(result, *args, **kw))
        try:
            if validate and not self.is_valid:
                data = json.repr(data)[:REPR_LIMIT]
                self.log.warning(f'{Who(cls)}: invalid data={data}')

        except ConfigurationError as e:

            for message in e.args[1].values():
                self.log.error(f'{Who(cls)}: {message}')  # noqa: TRY400

            self.log.error(str(e.args[0]))  # noqa: TRY400
            raise

        return self

    @property
    def extra_kwargs(self):
        return dict(self.__dict__['__non_model_items__'])

    @Property.Class.Cached
    def Defaults(cls):  # noqa: N802
        return {k: v.default for k, v in cls.__fields_dict__.items()}

    @Property.Mixed
    def as_dict(self) -> dict:
        return dict(self.__fields_dict__ if is_class(self) else self)

    @Property.Cached
    def as_json(self) -> str:
        return json.dumps(self.as_dict)

    @Property.Cached
    def as_sql(self) -> dict:
        def sql_cast(key, value):

            if value is None:
                return 'NULL'

            elif isinstance(value, str):
                if (
                    len(value) > 2 and  # noqa: PLR2004
                    value[0] == value[-1] and
                    value[0] in ('"', "'")
                ):
                    return value
                return f'"{value}"'

            elif isinstance(value, bool | int):
                return int(value)

            elif isinstance(value, datetime):
                return value.strftime('"%Y-%m-%d %H:%M:%S"')

            elif isinstance(value, date):
                return value.strftime('"%Y-%m-%d"')

            msg = f'{key=}: {Who.Is(value)}'
            self.log.exception(msg)
            raise TypeError(msg)

        return {k: sql_cast(k, v) for k, v in self.as_dict.items()}

    def to_sql(self, glue=',', /, callback=None):
        return glue.join(
            f'{k}={v}' for k, v in self.as_sql.items()
            if not is_callable(callback) or callback(v))

    @Property.Cached
    def export(self):
        return self.as_dict | self.extra_kwargs

    def copy(self, /, **kw):
        return self.load(self.export | kw)

    def __iter__(self):
        for key, field in self.__fields_dict__.items():
            try:
                value = getattr(self, key)

            except AttributeError as e:
                msg = (
                    f'{Who(self)}: declared ({Who(key)}){key} required, got {field=}')
                raise ConfigurationEvaluationError(msg) from e

            yield key, value

    @Property.Cached
    def fields(self):
        return tuple(map(itemgetter(0), self))

    def __and__(self, other):

        if self is other:
            return self.copy()

        elif isinstance(other, dataclass):
            other = other.export

        if isinstance(other, dict):
            return from_dict({
                k: v for k, v in self.export.items()
                if k in other and v == other[k]}, name=Who(other, full=False))

        elif is_collection(other):
            ...

        elif isinstance(other, str):
            other = [other]

        else:
            msg = f'{Who.Is(other)} must be dict | str | Iterable | {Who(dataclass)}'
            self.log.fatal(msg)
            raise TypeError(msg)

        data = self.export
        return from_dict(
            {k: data[k] for k in set(data) & set(other)}, name=Who(other, full=False))

    def __xor__(self, other):

        if not other:
            return self.copy()

        elif self is other:
            raise ValueError

        elif isinstance(other, dataclass):
            other = other.export

        if isinstance(other, dict):
            export = self.export
            return from_dict({
                k: export.get(k, other[k])
                for k, v in set(export) ^ set(other)}, name=Who(other, full=False))

        msg = f'{Who.Is(other)} must be dict | {Who(dataclass)}'
        self.log.fatal(msg)
        raise TypeError(msg)

    def __sub__(self, other):

        if not other:
            return self.copy()

        elif self is other:
            raise ValueError

        elif isinstance(other, dataclass):
            other = other.export

        if isinstance(other, dict):
            return self.load({
                k: v for k, v in self.export.items()
                if k not in other or v == other[k]})

        elif is_collection(other):
            ...

        elif isinstance(other, str):
            other = [other]

        else:
            msg = f'{Who.Is(other)} must be dict | str | Iterable | {Who(dataclass)}'
            self.log.fatal(msg)
            raise TypeError(msg)

        return from_dict(
            {k: v for k, v in self.export.items() if k not in other},
            name=Who(other, full=False))

    def __or__(self, other):

        if self is other or not other:
            return self.copy()

        elif isinstance(other, dataclass):
            other = other.export

        else:
            msg = f'{Who.Is(other)} must be dict | {Who(dataclass)}'
            self.log.fatal(msg)
            raise TypeError(msg)

        return from_dict(self.export | other, name=Who(other, full=False))

    def __add__(self, other):

        if isinstance(other, dataclass):
            other = other.export

        if isinstance(other, dict):
            return self.load({
                k: v for k, v in self.export.items()
                if k in other and v != other[k]})

        return self.__or__(other)

    @Property.Cached
    def is_valid(self):
        result = {}
        fields = self.__fields_dict__

        for key, value in self:
            field = fields[key]

            objecttype = field.type
            if isinstance(field.type, bool):
                objecttype = bool

            with suppress(TypeError):
                if Is.Class(objecttype) and isinstance(value, objecttype):
                    continue

            if not Is.subclass(value, objecttype):
                result[key] =(
                    f'({field.typename}){key} invalid type ({Who.Is(value)})')

        if result:
            msg = f"{Who(self)}: {'; '.join(sort(result))}"
            raise ConfigurationEvaluationError(msg, result)
        return True


class DataClassMutable(dataclass):
    __dataclass_options__ = {'frozen': False}  # noqa: RUF012


class DataClassFlexible(dataclass):
    policy = dataclass.Policy.Exclude


class BaseAutoClass(Logging.Mixin):
    __field__ = None

    @classmethod
    def __autoclasses_fields__(cls):
        hier = iter_inheritance(
            cls, exclude_stdlib=False,
            include=lambda x: Is.subclass(x, BaseAutoClass))

        for klass in hier:
            if klass is get_owner(klass, '__field__') and klass.__field__:
                yield klass

    @Property.Class.Parent
    def __autoclasses__(cls):
        return frozenset(map(attrgetter('__field__'), cls.__autoclasses_fields__()))


@cache
def make_auto(field):
    Field, _field = field.capitalize(), f'_{field}'  # noqa: N806

    class autoclass(BaseAutoClass):  # noqa: N801
        __field__ = field

        def __init__(self, *args, **kw) -> None:
            # TODO: валидировать (опционально) принимаемый конфиг проактивно здесь
            setattr(self, f'_{field}', kw.pop(field, Nothing))

            if (
                (owner := get_owner(BaseAutoClass, Field))
                and owner is get_owner(class_of(self), Field)
            ):
                self.log.warning(
                    f"{Who(self)}.{Field}{sourcefile(self, '(in %s)')} "
                    f"isn't defined, but required because "
                    f"it's {field} schema for instances")

            if (
                (owner := get_owner(BaseAutoClass, field))
                and owner != get_owner(class_of(self), field)
            ):
                self.log.warning(
                    f'{Who(self)}.{field} property overrided', once=True)

            elif getattr(self, f'_{field}', Nothing) is Nothing:
                self.log.warning(
                    f"{Who(self)}{sourcefile(self, 'from %s')} it's class "
                    f'with dataclass.BaseAutoClass (by {Who(BaseAutoClass)}), '
                    f'but initialized without {field}={{}} passed to __init__; '
                    f'{Who(self)}.{field} call now useless', trace=True, shift=-1)

            if self in args:
                where = 'as first arg' if args[0] is self else 'into args'
                self.log.warning(
                    f'{Who(self, addr=True)}: self {where} passed',
                    trace=True, shift=-1)

            try:
                super().__init__(*args, **kw)

            except TypeError:
                self.log.fatal(
                    f'{Who(self)}: {Who(super())}'
                    f'.__init__(*{args=}, **{kw=}) failed')
                raise

    def model(cls):
        msg = (
            f"{Who(cls)}.{Field}{sourcefile(cls, '(from %s)')} "
            f"isn't defined, but required because it's config schema for instances")
        raise TypeError(msg)

    def fill_model(self):
        scheme = getattr(self, Field)

        if scheme is None:
            msg = f"{Who(self)}.{Field} isn't defined or None instead dataclass"
            raise ConfigurationSchemeMissingError(msg)

        if not Is.subclass(scheme, dataclass):
            inheritance = ' -> '.join(map(Who, scheme.__mro_classes__))
            msg = f'{Who(self)}.{Field} must be dataclass, {inheritance=}'
            raise ConfigurationTypeError(msg)

        if getattr(self, _field, Nothing) is Nothing:
            classes = [
                f'{Who(cls)}{sourcefile(cls, "(%s)")}'
                for cls in scheme.__mro_classes__]
            msg = (
                f'{Who(self)}.{_field} missing, probably you forgot to pass '
                f'{field} dict as kwarg via super().__init__({field}=), check in '
                f"{', '.join(classes)}")
            raise ConfigurationSchemeMissingError(msg)

        config = getattr(self, _field)
        try:
            result = scheme.load(config)

        except Exception as e:
            config = json.repr(config)
            inheritance = ' -> '.join(map(Who, scheme.__mro_classes__))
            msg = (
                f"{Who(self)} couldn't load {config} "
                f'into {scheme=} ({inheritance})')
            if getattr(self, _field, None) is None:
                msg = (
                    f'{msg}, maybe you forgot to pass {field} as kwarg to '
                    f'{Who(self, full=False)}({field}=)?')
            raise ConfigurationSchemeError(msg) from e

        self.log.debug(f'result={result}', once=False)
        return result

    autoclass.__name__ = f'{autoclass.__name__}({field})'
    autoclass.__qualname__ = str(autoclass.__name__)

    model.__name__ = Field
    fill_model.__name__ = field

    model.__qualname__ = f'{autoclass.__qualname__}.{Field}'
    fill_model.__qualname__ = f'{autoclass.__qualname__}.{field}'

    setattr(autoclass, Field, Property.Class.Cached(model))
    setattr(autoclass, field, Property.Cached(fill_model))

    return autoclass


def autoclass(*args):
    if len(args) == 1:
        return make_auto(args[0])
    return tuple(map(make_auto, args))


def simple(classname, *names, **kw):
    mro = kw.pop('mro', None) or ()
    if not is_iterable(mro):
        mro = [mro]
    mro = [*list(mro), BaseAutoClass, dataclass]

    fields = {}
    default = _MISSING_TYPE()
    for name in names:
        field = Field(default, default, True, True, None, True, {}, True)  # noqa: FBT003
        field.name = name
        field.type = Optional[Any]  # noqa: UP007
        field._field_type = _FIELD  # noqa: SLF001
        fields[name] = field

    fields['__annotations__'] = {f.name: f.type for f in fields.values()}

    return type(classname, tuple(mro), fields)


def from_dict(*args, **kw):
    mro = kw.pop('mro', [])

    if (
        len(args) >= 2 and  # noqa: PLR2004
        isinstance(args[0], str) and
        isinstance(args[1], dict)
    ):
        classname, data = args[:2]
        args = args[2:]

    elif (
        len(args) == 1 and kw and
        isinstance(args[0], str)
    ):
        classname = args[0]
        args, data = (), kw.copy()

    elif (
        len(args) == 1 and
        kw.get('name') and
        isinstance(args[0], dict)
    ):
        classname = kw.pop('name')
        args, data = args[1:], args[0]

    elif args and isinstance(args[0], dict):
        classname, data = Who(autoclass), args[0]
        args = args[1:]

    elif not args and kw:
        classname, data = Who(autoclass), kw.copy()

    else:
        msg = f'for {Who(autoclass)}(*{args=}, **{kw=})'
        raise NotImplementedError(msg)

    return simple(classname, *data, mro=mro).load(data, *args, **kw)


def from_tuple(*fields, **kw):

    def wrapper(*args, **kwlocal):

        if len(args) == 1 and is_collection(args):
            args = args[0]

        if len(args) != len(fields):
            msg = f'{Who.Is(args)} must be {len(fields)} args: {fields}'
            raise TypeError(msg)

        return from_dict(dict(zip(fields, args, strict=False)), **(kw | kwlocal))

    return wrapper


def from_module(module, /, order=None, name=None):

    if isinstance(module, bytes | str):
        module = required(Str.to_ascii(module))

    if not isinstance(module, ModuleType):
        msg = f'{Who.Is(module)} must be {Who(ModuleType)} or bytes | str'
        Logging.get(module).fatal(msg)
        raise TypeError(msg)

    data = {
        name: getattr(module, name)
        for name in (order or module.__all__)}
    return dataclass.FromDict(name, data) if name else dataclass.FromDict(data)


def from_namespace(namespace, /, callback=capitalize_snakes, name=None):
    data = {(callback(k) if callback else k): v for k, v in namespace._get_kwargs()}  # noqa: SLF001
    return dataclass.FromDict(name, data) if name else dataclass.FromDict(data)


dataclass.base      = BaseAutoClass
dataclass.exception = ConfigurationError

dataclass.flex      = DataClassFlexible
dataclass.mutable   = DataClassMutable

dataclass.auto      = autoclass
dataclass.config    = autoclass('config')
dataclass.simple    = simple
dataclass.dict      = from_dict
dataclass.module    = from_module
dataclass.space     = from_namespace
dataclass.tuple     = from_tuple



import atexit
import signal
import sys
import threading
import time
import warnings
from collections.abc import Callable
from pathlib import Path
from signal import signal as bind
from types import FrameType
from typing import Any

from kalib.classes import Singleton
from kalib.descriptors import cache
from kalib.internals import Who
from kalib.loggers import Logging

Logger = Logging.Default
NeedRestart = False


@cache
def get_selfpath():
    return Path(sys.argv[0]).resolve()


def get_mtime():
    return get_selfpath().stat().st_mtime


@cache
def quit_at(*, func=sys.exit, signal=None, errno=137, **kw):
    """Quit the program when the runned file is updated or signal is received."""

    def handler(*_):
        global NeedRestart  # noqa: PLW0603
        NeedRestart = True
        Logger.warning(f'{signal=} received, mark restart as needed')

    if signal:
        bind(signal, handler)

    initial_stamp = get_mtime()

    def on_change(*, sleep=0.0):

        if NeedRestart and signal:
            Logger.warning(f'{signal=} received, quit..')
            func(errno)
            return False

        try:
            if initial_stamp != (ctime := get_mtime()):
                file = str(get_selfpath())
                Logger.warning(
                    f'{file=} updated {time.time() - ctime:.2f} seconds ago, quit..')
                func(errno)
                return False

        except FileNotFoundError:
            Logger.warning(
                f'{get_selfpath()} not found, quit..')
            return False

        if sleep := (sleep or kw.get('sleep', 0.0)):
            time.sleep(sleep)
        return True
    return on_change


class OnExit(metaclass=Singleton):

    def __init__(self):

        self.callbacks   = []
        self.hooks_chain = []

        self.original_hook  = sys.excepthook
        self.already_called = False

        self.inject_hook()
        self.inject_signal_handler()
        self.inject_threading_hook()

        atexit.register(self.teardown)

    def inject_hook(self) -> None:
        sys.excepthook = self.exceptions_hooks_proxy

    def exceptions_hooks_proxy(
        self,
        exc_type  : type[BaseException],
        exc_value : BaseException,
        traceback : Any,
    ) -> None:

        if sys.excepthook != self.exceptions_hooks_proxy:
            self.hooks_chain.append(sys.excepthook)
            self.inject_hook()

        for hook in (*self.hooks_chain, self.original_hook):
            try:
                hook(exc_type, exc_value, traceback)
            except Exception as e:  # noqa: BLE001
                warnings.warn(f'{Who(hook)}: {e!r}', RuntimeWarning, stacklevel=2)

        self.teardown()

    def inject_signal_handler(self) -> None:
        for sig in signal.SIGINT, signal.SIGTERM, signal.SIGQUIT:
            bind(sig, self.signal_handler)

    def signal_handler(self, _: int, frame: FrameType | None) -> None:
        self.teardown()
        sys.exit(1)

    def inject_threading_hook(self) -> None:
        threading.excepthook = self.threading_handler

    def threading_handler(self, args: threading.ExceptHookArgs) -> None:
        if args.exc_type is SystemExit:
            return

        self.exceptions_hooks_proxy(
            args.exc_type,
            args.exc_value,
            args.exc_traceback)

    def restore_original_handlers(self) -> None:
        bind(signal.SIGINT, signal.SIG_DFL)
        bind(signal.SIGTERM, signal.SIG_DFL)
        bind(signal.SIGHUP, signal.SIG_DFL)

        sys.excepthook = self.original_hook
        threading.excepthook = threading.__excepthook__

    #

    def schedule(self, func: Callable) -> None:
        self.callbacks.append(func)

    def add_hook(self, func: Callable) -> None:
        self.hooks_chain.append(func)

    #

    def teardown(self) -> None:
        if self.already_called:
            return

        try:
            for func in self.callbacks:
                try:
                    func()
                except BaseException as e:  # noqa: BLE001
                    warnings.warn(f'{Who(func)}: {e!r}', RuntimeWarning, stacklevel=2)

        finally:
            self.already_called = True
            self.restore_original_handlers()


Exit = OnExit()



import sys
from datetime import datetime, timedelta, timezone
from functools import partial
from itertools import cycle
from math import log2
from operator import attrgetter, methodcaller
from pathlib import Path
from random import shuffle
from sys import stderr, stdin, stdout, version_info
from time import monotonic, time

from kalib.classes import Missing
from kalib.descriptors import Property, cache
from kalib.functions import to_bytes
from kalib.internals import (
    Who,
    class_of,
    get_attr,
    is_class,
    is_iterable,
)
from kalib.internals import sourcefile as generic_sourcefile

Nothing = Missing()


@cache
def get_importer():
    # avoid circular imports
    from kalib.importer import required
    return required


@cache
def get_toml_loader():
    if version_info >= (3, 11):
        from tomllib import load
    else:
        try:
            from tomli import load
        except ImportError as e:
            raise ImportError('install "tomli" or upgrade to Python 3.11+') from e
    return load


@cache
def get_yaml_loader():
    return get_importer()('yaml.safe_load')


def is_tty():
    """Return True if all standard streams are tty."""

    if not getattr(sys, 'frozen', False):  # nuitka compiler checks this
        return all(map(methodcaller('isatty'), (stderr, stdin, stdout)))


tty = is_tty()


def Now(tz=True, /, **kw):  # noqa: N802
    """Return datetime object with or without timezone.

    When tz passed, it's used, when not tz — timezone-naive datetime returned.

    Also keywargs can be passed to timedelta, so you can get datetime in the future.
    """

    if tz:
        result = datetime.now(tz=timezone.utc if tz is True else tz)
    else:
        result = datetime.utcnow()  # noqa: DTZ003

    return (result + timedelta(**kw)) if kw else result


def stamp(**kw):
    """Return timestamp with timedelta keywargs with timezone support."""
    return Now(**kw).replace(tzinfo=None).timestamp()


class Timer:
    """Context manager to measure time spent in block."""

    @Property.Class
    def now(cls):
        return time()

    @Property.Class
    def stamp(cls):
        return stamp()

    def __init__(self):
        self._start = None
        self.spent = None

    def __enter__(self):
        self._start = monotonic()

    def __exit__(self, *args):
        self.spent = monotonic() - self._start

    @property
    def delta(self):
        return monotonic() - self._start if self._start else 0.0


def lazy_proxy_to(  # noqa: PLR0915
    *mapping,
    getter  = attrgetter,
    default = Nothing,
    pre     = None,
    safe    = True,
):
    if isinstance(mapping[-1], str):
        bind = Property.Cached

    elif mapping[-1] is None:
        bind, mapping = None, mapping[:-1]

    else:
        bind, mapping = mapping[-1], mapping[:-1]

    def binder(cls):  # noqa: PLR0915

        try:
            fields = cls.__proxy_fields__
        except AttributeError:
            fields = []
            cls.__proxy_fields__ = fields

        pivot, mapping_list = mapping[0], mapping[1:]

        if not is_class(cls):
            msg = f"{Who.Is(cls)} isn't a class"
            raise TypeError(msg)

        if (
            not mapping_list or
            (len(mapping_list) == 1 and not isinstance(mapping_list[0], str))
        ):
            raise ValueError(f'empty {mapping_list=} for {pivot=}')

        for method in mapping_list:

            if safe and not method.startswith('_') and get_attr(cls, method):
                msg = (
                    f'{Who(cls)} already exists {method!a}: '
                    f'{get_attr(cls, method)}')
                raise TypeError(msg)

            def lazy_call(method, node):
                if not isinstance(pivot, str):
                    try:
                        return getattr(pivot, method)
                    except AttributeError as e:
                        msg = (
                            f'{Who(node)}.{method} {Who.Name(getter)[:4]}-proxied -> '
                            f"{Who(pivot)}.{method}, but last isn't exists")
                        raise TypeError(msg) from e

                try:
                    entity = getattr(node, pivot)
                except AttributeError as e:
                    msg = (
                        f'{Who(node)}.{method} {Who.Name(getter)[:4]}-proxied -> '
                        f'{Who(node)}.{pivot}.{method}, but '
                        f"{Who(node)}.{pivot} isn't exists")
                    raise TypeError(msg) from e

                if entity is None:
                    msg = (
                        f'{Who(node)}.{method} {Who.Name(getter)[:4]}-proxied -> '
                        f'{Who(node)}.{pivot}.{method}, but current '
                        f'{Who(node)}.{pivot} is None')

                    if default is Nothing:
                        raise TypeError(msg)

                    msg = f'{msg}; return {Who.Is(default)}'
                    cls.log.verbose(msg)
                    attribute = default

                else:
                    try:
                        attribute = getter(method)(entity)

                    except (AttributeError, KeyError) as e:
                        msg = (
                            f'{Who(node)}.{method} {Who.Name(getter)[:4]}-proxied -> '
                            f"{Who(node)}.{pivot}.{method}, but isn't exists "
                            f"('{method}' not in {Who(node)}.{pivot}): {Who.Is(entity)}")

                        if default is Nothing:
                            raise class_of(e)(msg) from e

                        msg = f'{msg}; return {Who.Is(default)}'
                        cls.log.verbose(msg)
                        attribute = default

                return partial(pre, attribute) if pre else attribute

            lazy_call.__name__ = method
            lazy_call.__qualname__ = f'{pivot}.{method}'

            if bind is None:
                node = cls.__dict__[pivot]
                try:
                    value = node.__dict__[method]
                except KeyError:
                    value = getattr(node, method)
            else:
                wrap = partial(lazy_call, method)
                wrap.__name__ = method
                wrap.__qualname__ = f'{pivot}.{method}'
                value = bind(wrap)

            fields.append(method)
            setattr(cls, method, value)
            cls.__proxy_fields__.sort()

        return cls
    return binder


def sourcefile(something, template=None):
    """Return sourcefile for object if it's reachable with format template"""
    source = generic_sourcefile(something)
    if not source:
        return ''
    elif not template:
        return str(source)
    return ' ' + (template % source)


def to_tuple(x):
    return tuple(x or ()) if is_iterable(x) else (x,)


def toml_read(path):
    loader = get_toml_loader()
    with Path(path).open('rb') as fd:
        return loader(fd)


def yaml_read(path):
    loader = get_yaml_loader()
    with Path(path).open('rb') as fd:
        return loader(fd.read())


def cycler(array):
    """Return a function that cycles through the given array."""
    iterator = cycle(array)
    return lambda: next(iterator)


def shuffler(iterable):
    """Return a function that shuffles the given array."""

    if len(iterable) < 3:  # noqa: PLR2004
        return cycler(iterable)

    def iterator():
        last = None
        while True:
            array = list(iterable)
            shuffle(array)

            if last and array[0] is last[-1]:
                continue

            yield from array
            last = array

    iterator = iterator()
    return lambda: next(iterator)


def set_title(title=None, short=True):

    opts = 0
    func = get_importer()('setproctitle.setproctitle')

    if title:
        name = title
    else:
        script = sys.argv[0]
        if script.endswith('.py'):
            opts = 1
            name = Path(script).stem
        else:
            name = Path(sys.executable).stem

    func(f"{name} {'' if short else ' '.join(sys.argv[opts:])}".strip())


@cache(2 ** 10)
def fasthash(path, algo='xxhash.xxh128'):
    """Generate fast cache by file content according to sector size."""

    hasher = get_importer()(algo)()

    path = Path(path)
    size = path.stat().st_size

    unit_power = 12  # cluster power, modern is 12, 4096 bytes
    cluster_size = 2 ** unit_power  # cluster size in bytes

    parts = int(log2(size))  # file parts count
    part_size = int(size / parts)  # every part size
    cluster = 2 ** min(12, int(log2(part_size)))  # read block size

    with path.open('rb') as fd:
        hasher.update(to_bytes(str(size)))
        for no in range(parts):

            # seek to every file part cluster start and read cluster
            fd.seek(int((part_size * no) // cluster_size) * cluster_size)
            hasher.update(to_bytes(fd.read(cluster)))

    return hasher.hexdigest()


def is_python_runtime():
    """Checks if current runtime is Python, not Nuitka or PyInstaller."""
    return Path(sys.executable).is_file()



from contextlib import suppress
from os import walk
from pathlib import Path
from re import match

from kalib.datastructures import json
from kalib.descriptors import Property, cache
from kalib.importer import optional, required, sort
from kalib.internals import Who, iter_stack, sourcefile
from kalib.loggers import Logging
from kalib.misc import is_python_runtime

DEFAULT_VERSION_RE = r'^(v\d+\.py)$'

logger = Logging.Default


class Git(Logging.Mixin):

    @Property.Class.Cached
    def Repo(cls):  # noqa: N802
        return required('git.Repo')

    @Property.Class.Parent
    def root(cls):
        expect = suppress(required('git.exc.InvalidGitRepositoryError', quiet=True))
        with expect, cls.Repo(
            Path.cwd().resolve(),
            search_parent_directories=True,
        ) as repo:
            return Path(repo.working_tree_dir).resolve()
        cls.log.verbose('not detected')

    @Property.Class.Parent
    def repo(cls):
        if (path := cls.root):
            return cls.Repo(path, search_parent_directories=True)
        cls.log.verbose('not found')

    @Property.Class.Parent
    def from_path(cls):
        root = cls.root

        if not root:
            cls.log.warning("feature disabled, git root isn't detected")
            return (lambda *args: None)  # noqa: ARG005

        @cache
        def resolver(path):
            for commit in cls.repo.iter_commits(paths=path, max_count=1):
                return commit
            cls.log.warning(f'{path=} not in repository {root!s} index')

        @cache
        def explorer(path):
            file = Path(path).resolve()

            if not file.is_file():
                cls.log.warning(f'{path=} ({file!s}) not found')
                return

            elif not str(file).startswith(str(root)):
                cls.log.warning(f'{path=} ({file!s}) not in {root!s}')
                return

            return resolver(str(file)[len(str(root)) + 1:])

        return explorer

    @Property.Class.Parent
    def files(cls):
        result = []
        prefix = len(str(cls.root)) + 1
        for rt, _, fs in walk(cls.root):
            root = Path(rt)
            for f in fs:
                if f.endswith('.py'):
                    result.append(str(root / f)[prefix:])

        return tuple(sort(result))

    @Property.Class.Parent
    def tree(cls):
        return {
            path: str(cls.from_path(path)) for path in cls.files
            if cls.from_path(path) is not None}

    @Property.Class.Parent
    def tag(cls):
        with suppress(Exception), cls.repo as git:
            head = git.head.commit

            def get_distance(tag):
                return int(cli.rev_list(
                    '--count', f'{tag.commit.hexsha}..{head.hexsha}'))

            if git.tags:
                cli = required('git.Git')()
                distances = {
                    tag: get_distance(tag)
                    for tag in sort(git.tags, key=lambda x: x.name, reverse=True)}

                if distances:
                    return min(distances, key=distances.get)

    @Property.Class.Parent
    def version(cls):
        with suppress(ImportError):
            if cls.tag and (version := optional('versioneer.get_version')):
                return version()


@cache
def latest(name, path, pattern=None):
    files = tuple(filter(
        lambda x: match(pattern or DEFAULT_VERSION_RE, x.name),
        Path(path).parent.glob('*.py')))

    if files:
        version = Path(sort(files)[-1]).stem
        module = getattr(__import__(name, globals(), locals(), [version]), version)
        logger.verbose(f'{name} {module=}')
        return module


def add_versioned_path(pattern=None):
    """Add the latest versioned module to the caller's namespace."""

    if not is_python_runtime():
        raise NotImplementedError('this function is for python runtime only')

    def get_call_frame():
        """Get the frame of the caller of the function that called this
        function."""
        for frame in iter_stack(offset=2):
            if frame.filename == __file__:  # skip self
                continue
            if frame.code_context is not None:
                return frame
        raise IndexError

    try:
        pointer = get_call_frame()

    except IndexError:
        raise RuntimeError('could not find the caller frame') from None

    local  = pointer.frame.f_locals
    module = latest(local['__name__'], pointer.filename, pattern)

    to_import = getattr(module, '__all__', ())
    if isinstance(to_import, str):
        # we can override and use as __all__ == 'EntityName'
        to_import = (to_import,)

    local['latest'] = module
    for something in to_import:
        try:
            local[something] = getattr(module, something)

        except AttributeError:

            msg = (
                f"couldn't import {something!a} from module {Who.Name(module)!a}; "
                f"may be it's not found, but declared in {Who.Name(module)}.__all__ = "
                f'{json.repr(to_import)} in {sourcefile(module)!r}')

            logger.critical(msg)
            raise ImportError(msg) from None

    return module
